@@ -15,7 +15,7 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
- * Copyright (C) 2005 - 2008 Red Hat, Inc.
+ * Copyright (C) 2005 - 2010 Red Hat, Inc.
  * Copyright (C) 2006 - 2008 Novell, Inc.
  */
 
@@ -35,7 +35,8 @@
 #include "nm-setting-ip4-config.h"
 #include "nm-dbus-manager.h"
 #include "nm-manager.h"
-#include "NetworkManagerSystem.h"
+#include "nm-system.h"
+#include "nm-logging.h"
 #include "nm-utils.h"
 #include "nm-vpn-plugin-bindings.h"
 #include "nm-marshal.h"
@@ -43,18 +44,22 @@
 #include "nm-properties-changed-signal.h"
 #include "nm-dbus-glib-types.h"
 #include "NetworkManagerUtils.h"
-#include "nm-named-manager.h"
-#include "nm-netlink.h"
+#include "nm-dns-manager.h"
+#include "nm-netlink-monitor.h"
+#include "nm-glib-compat.h"
 
 #include "nm-vpn-connection-glue.h"
 
-G_DEFINE_TYPE (NMVPNConnection, nm_vpn_connection, G_TYPE_OBJECT)
+static void secrets_provider_interface_init (NMSecretsProviderInterface *sp_interface_class);
+
+G_DEFINE_TYPE_EXTENDED (NMVPNConnection, nm_vpn_connection, G_TYPE_OBJECT, 0,
+                        G_IMPLEMENT_INTERFACE (NM_TYPE_SECRETS_PROVIDER_INTERFACE,
+                                               secrets_provider_interface_init))
 
 typedef struct {
 	gboolean disposed;
 
 	NMConnection *connection;
-	DBusGProxyCall *secrets_call;
 
 	NMActRequest *act_request;
 	char *ac_path;
@@ -64,6 +69,7 @@ typedef struct {
 	gulong device_ip4;
 
 	gboolean is_default;
+	gboolean is_default6;
 	NMActiveConnectionState state;
 
 	NMVPNConnectionState vpn_state;
@@ -72,7 +78,7 @@ typedef struct {
 	guint ipconfig_timeout;
 	NMIP4Config *ip4_config;
 	guint32 ip4_internal_gw;
-	char *tundev;
+	char *ip_iface;
 	char *banner;
 
 	struct rtnl_route *gw_route;
@@ -97,6 +103,7 @@ enum {
 	PROP_DEVICES,
 	PROP_STATE,
 	PROP_DEFAULT,
+	PROP_DEFAULT6,
 	PROP_VPN,
 	PROP_VPN_STATE,
 	PROP_BANNER,
@@ -112,6 +119,7 @@ nm_vpn_connection_set_vpn_state (NMVPNConnection *connection,
 	NMVPNConnectionPrivate *priv;
 	NMActiveConnectionState new_ac_state;
 	NMVPNConnectionState old_vpn_state;
+	char *ip_iface;
 
 	g_return_if_fail (NM_IS_VPN_CONNECTION (connection));
 
@@ -123,6 +131,11 @@ nm_vpn_connection_set_vpn_state (NMVPNConnection *connection,
 	old_vpn_state = priv->vpn_state;
 	priv->vpn_state = vpn_state;
 
+	/* Save ip_iface since when the VPN goes down it may get freed
+	 * before we're done with it.
+	 */
+	ip_iface = g_strdup (priv->ip_iface);
+
 	/* Set the NMActiveConnection state based on VPN state */
 	switch (vpn_state) {
 	case NM_VPN_CONNECTION_STATE_PREPARE:
@@ -159,7 +172,7 @@ nm_vpn_connection_set_vpn_state (NMVPNConnection *connection,
 		nm_utils_call_dispatcher ("vpn-up",
 		                          priv->connection,
 		                          priv->parent_dev,
-		                          priv->tundev);
+		                          ip_iface);
 		break;
 	case NM_VPN_CONNECTION_STATE_FAILED:
 	case NM_VPN_CONNECTION_STATE_DISCONNECTED:
@@ -167,13 +180,14 @@ nm_vpn_connection_set_vpn_state (NMVPNConnection *connection,
 			nm_utils_call_dispatcher ("vpn-down",
 			                          priv->connection,
 			                          priv->parent_dev,
-			                          priv->tundev);
+			                          ip_iface);
 		}
 		break;
 	default:
 		break;
 	}
 
+	g_free (ip_iface);
 	g_object_unref (connection);
 }
 
@@ -205,7 +219,8 @@ device_ip4_config_changed (NMDevice *device,
 	NMVPNConnection *vpn = NM_VPN_CONNECTION (user_data);
 	NMVPNConnectionPrivate *priv = NM_VPN_CONNECTION_GET_PRIVATE (vpn);
 
-	if (priv->vpn_state != NM_VPN_CONNECTION_STATE_ACTIVATED)
+	if (   (priv->vpn_state != NM_VPN_CONNECTION_STATE_ACTIVATED)
+	    || !nm_device_get_ip4_config (device))
 		return;
 
 	if (priv->gw_route)
@@ -220,31 +235,31 @@ nm_vpn_connection_new (NMConnection *connection,
                        NMActRequest *act_request,
                        NMDevice *parent_device)
 {
-	NMVPNConnection *vpn_connection;
+	NMVPNConnection *self;
 	NMVPNConnectionPrivate *priv;
 
 	g_return_val_if_fail (NM_IS_CONNECTION (connection), NULL);
 	g_return_val_if_fail (NM_IS_ACT_REQUEST (act_request), NULL);
 	g_return_val_if_fail (NM_IS_DEVICE (parent_device), NULL);
 
-	vpn_connection = (NMVPNConnection *) g_object_new (NM_TYPE_VPN_CONNECTION, NULL);
-	if (!vpn_connection)
+	self = (NMVPNConnection *) g_object_new (NM_TYPE_VPN_CONNECTION, NULL);
+	if (!self)
 		return NULL;
 
-	priv = NM_VPN_CONNECTION_GET_PRIVATE (vpn_connection);
+	priv = NM_VPN_CONNECTION_GET_PRIVATE (self);
 
 	priv->connection = g_object_ref (connection);
-	priv->act_request = g_object_ref (act_request);
 	priv->parent_dev = g_object_ref (parent_device);
+	priv->act_request = g_object_ref (act_request);
 
 	priv->device_monitor = g_signal_connect (parent_device, "state-changed",
 									 G_CALLBACK (device_state_changed),
-									 vpn_connection);
+									 self);
 
 	priv->device_ip4 = g_signal_connect (parent_device, "notify::" NM_DEVICE_INTERFACE_IP4_CONFIG,
 	                                     G_CALLBACK (device_ip4_config_changed),
-	                                     vpn_connection);
-	return vpn_connection;
+	                                     self);
+	return self;
 }
 
 static const char *
@@ -264,7 +279,7 @@ plugin_failed (DBusGProxy *proxy,
 {
 	NMVPNConnectionPrivate *priv = NM_VPN_CONNECTION_GET_PRIVATE (user_data);
 
-	nm_info ("VPN plugin failed: %d", plugin_failure);
+	nm_log_warn (LOGD_VPN, "VPN plugin failed: %d", plugin_failure);
 
 	switch (plugin_failure) {
 	case NM_VPN_PLUGIN_FAILURE_LOGIN_FAILED:
@@ -280,39 +295,38 @@ plugin_failed (DBusGProxy *proxy,
 
 static void
 plugin_state_changed (DBusGProxy *proxy,
-				  NMVPNServiceState state,
-				  gpointer user_data)
+                      NMVPNServiceState state,
+                      gpointer user_data)
 {
 	NMVPNConnection *connection = NM_VPN_CONNECTION (user_data);
 	NMVPNConnectionPrivate *priv = NM_VPN_CONNECTION_GET_PRIVATE (connection);
 
-	nm_info ("VPN plugin state changed: %d", state);
+	nm_log_info (LOGD_VPN, "VPN plugin state changed: %d", state);
 
-	if (state != NM_VPN_SERVICE_STATE_STOPPED)
-		return;
+	if (state == NM_VPN_SERVICE_STATE_STOPPED) {
+		/* Clear connection secrets to ensure secrets get requested each time the
+		 * connection is activated.
+		 */
+		nm_connection_clear_secrets (priv->connection);
 
-	switch (nm_vpn_connection_get_vpn_state (connection)) {
-	case NM_VPN_CONNECTION_STATE_PREPARE:
-	case NM_VPN_CONNECTION_STATE_NEED_AUTH:
-	case NM_VPN_CONNECTION_STATE_CONNECT:
-	case NM_VPN_CONNECTION_STATE_IP_CONFIG_GET:
-	case NM_VPN_CONNECTION_STATE_ACTIVATED:
-		nm_info ("VPN plugin state change reason: %d", priv->failure_reason);
-		nm_vpn_connection_set_vpn_state (connection,
-		                                 NM_VPN_CONNECTION_STATE_FAILED,
-										 priv->failure_reason);
+		switch (nm_vpn_connection_get_vpn_state (connection)) {
+		case NM_VPN_CONNECTION_STATE_PREPARE:
+		case NM_VPN_CONNECTION_STATE_NEED_AUTH:
+		case NM_VPN_CONNECTION_STATE_CONNECT:
+		case NM_VPN_CONNECTION_STATE_IP_CONFIG_GET:
+		case NM_VPN_CONNECTION_STATE_ACTIVATED:
+			nm_log_info (LOGD_VPN, "VPN plugin state change reason: %d", priv->failure_reason);
+			nm_vpn_connection_set_vpn_state (connection,
+			                                 NM_VPN_CONNECTION_STATE_FAILED,
+											 priv->failure_reason);
 
-		/* Reset the failure reason */
-		priv->failure_reason = NM_VPN_CONNECTION_STATE_REASON_UNKNOWN;
-		break;
-	default:
-		break;
+			/* Reset the failure reason */
+			priv->failure_reason = NM_VPN_CONNECTION_STATE_REASON_UNKNOWN;
+			break;
+		default:
+			break;
+		}
 	}
-
-	/* Clear connection secrets too so the auth dialogs get asked
-	 * for them next time.
-	 */
-	nm_connection_clear_secrets (priv->connection);
 }
 
 static const char *
@@ -327,8 +341,8 @@ ip_address_to_string (guint32 numeric)
 	if (inet_ntop (AF_INET, &temp_addr, buf, INET_ADDRSTRLEN)) {
 		return buf;
 	} else {
-		nm_warning ("%s: error converting IP4 address 0x%X",
-		            __func__, ntohl (temp_addr.s_addr));
+		nm_log_warn (LOGD_VPN, "error converting IP4 address 0x%X",
+		             ntohl (temp_addr.s_addr));
 		return NULL;
 	}
 }
@@ -336,7 +350,7 @@ ip_address_to_string (guint32 numeric)
 static void
 print_vpn_config (NMIP4Config *config,
                   guint32 internal_gw,
-                  const char *tundev,
+                  const char *ip_iface,
                   const char *banner)
 {
 	NMIP4Address *addr;
@@ -347,38 +361,47 @@ print_vpn_config (NMIP4Config *config,
 
 	addr = nm_ip4_config_get_address (config, 0);
 
-	nm_info ("VPN Gateway: %s", ip_address_to_string (nm_ip4_address_get_gateway (addr)));
+	nm_log_info (LOGD_VPN, "VPN Gateway: %s", ip_address_to_string (nm_ip4_address_get_gateway (addr)));
 	if (internal_gw)
-		nm_info ("Internal Gateway: %s", ip_address_to_string (internal_gw));
-	nm_info ("Tunnel Device: %s", tundev);
-	nm_info ("Internal IP4 Address: %s", ip_address_to_string (nm_ip4_address_get_address (addr)));
-	nm_info ("Internal IP4 Prefix: %d", nm_ip4_address_get_prefix (addr));
-	nm_info ("Internal IP4 Point-to-Point Address: %s",
-		    ip_address_to_string (nm_ip4_config_get_ptp_address (config)));
-	nm_info ("Maximum Segment Size (MSS): %d", nm_ip4_config_get_mss (config));
+		nm_log_info (LOGD_VPN, "Internal Gateway: %s", ip_address_to_string (internal_gw));
+	nm_log_info (LOGD_VPN, "Tunnel Device: %s", ip_iface);
+	nm_log_info (LOGD_VPN, "Internal IP4 Address: %s", ip_address_to_string (nm_ip4_address_get_address (addr)));
+	nm_log_info (LOGD_VPN, "Internal IP4 Prefix: %d", nm_ip4_address_get_prefix (addr));
+	nm_log_info (LOGD_VPN, "Internal IP4 Point-to-Point Address: %s",
+	             ip_address_to_string (nm_ip4_config_get_ptp_address (config)));
+	nm_log_info (LOGD_VPN, "Maximum Segment Size (MSS): %d", nm_ip4_config_get_mss (config));
 
 	num = nm_ip4_config_get_num_routes (config);
 	for (i = 0; i < num; i++) {
 		NMIP4Route *route;
 
 		route = nm_ip4_config_get_route (config, i);
-		nm_info ("Static Route: %s/%d   Next Hop: %s",
-			    ip_address_to_string (nm_ip4_route_get_dest (route)),
-			    nm_ip4_route_get_prefix (route),
-			    ip_address_to_string (nm_ip4_route_get_next_hop (route)));
+		nm_log_info (LOGD_VPN, "Static Route: %s/%d   Next Hop: %s",
+		             ip_address_to_string (nm_ip4_route_get_dest (route)),
+		             nm_ip4_route_get_prefix (route),
+		             ip_address_to_string (nm_ip4_route_get_next_hop (route)));
 	}
 
+	nm_log_info (LOGD_VPN, "Forbid Default Route: %s",
+	             nm_ip4_config_get_never_default (config) ? "yes" : "no");
+
 	num = nm_ip4_config_get_num_nameservers (config);
-	for (i = 0; i < num; i++)
-		nm_info ("Internal IP4 DNS: %s", ip_address_to_string (nm_ip4_config_get_nameserver (config, i)));
+	for (i = 0; i < num; i++) {
+		nm_log_info (LOGD_VPN, "Internal IP4 DNS: %s",
+		             ip_address_to_string (nm_ip4_config_get_nameserver (config, i)));
+	}
 
 	if (nm_ip4_config_get_num_domains (config) > 0)
 		dns_domain = (char *) nm_ip4_config_get_domain (config, 0);
-	nm_info ("DNS Domain: '%s'", dns_domain ? dns_domain : "(none)");
-	nm_info ("Login Banner:");
-	nm_info ("-----------------------------------------");
-	nm_info ("%s", banner);
-	nm_info ("-----------------------------------------");
+
+	nm_log_info (LOGD_VPN, "DNS Domain: '%s'", dns_domain ? dns_domain : "(none)");
+
+	if (banner && strlen (banner)) {
+		nm_log_info (LOGD_VPN, "Login Banner:");
+		nm_log_info (LOGD_VPN, "-----------------------------------------");
+		nm_log_info (LOGD_VPN, "%s", banner);
+		nm_log_info (LOGD_VPN, "-----------------------------------------");
+	}
 }
 
 static void
@@ -393,9 +416,10 @@ nm_vpn_connection_ip4_config_get (DBusGProxy *proxy,
 	NMIP4Config *config;
 	GValue *val;
 	int i;
+	guint32 vpn_ext_gw = 0;
 
-	nm_info ("VPN connection '%s' (IP Config Get) reply received.",
-		    nm_vpn_connection_get_name (connection));
+	nm_log_info (LOGD_VPN, "VPN connection '%s' (IP Config Get) reply received.",
+	             nm_vpn_connection_get_name (connection));
 
 	g_source_remove (priv->ipconfig_timeout);
 	priv->ipconfig_timeout = 0;
@@ -404,9 +428,9 @@ nm_vpn_connection_ip4_config_get (DBusGProxy *proxy,
 
 	val = (GValue *) g_hash_table_lookup (config_hash, NM_VPN_PLUGIN_IP4_CONFIG_TUNDEV);
 	if (val)
-		priv->tundev = g_strdup (g_value_get_string (val));
+		priv->ip_iface = g_strdup (g_value_get_string (val));
 	else {
-		nm_warning ("%s: invalid or missing tunnel device received!", __func__);
+		nm_log_err (LOGD_VPN, "invalid or missing tunnel device received!");
 		goto error;
 	}
 
@@ -420,8 +444,10 @@ nm_vpn_connection_ip4_config_get (DBusGProxy *proxy,
 
 	/* External world-visible address of the VPN server */
 	val = (GValue *) g_hash_table_lookup (config_hash, NM_VPN_PLUGIN_IP4_CONFIG_EXT_GATEWAY);
-	if (val)
+	if (val) {
 		nm_ip4_address_set_gateway (addr, g_value_get_uint (val));
+		vpn_ext_gw = g_value_get_uint (val);
+	}
 
 	val = (GValue *) g_hash_table_lookup (config_hash, NM_VPN_PLUGIN_IP4_CONFIG_ADDRESS);
 	if (val)
@@ -438,7 +464,7 @@ nm_vpn_connection_ip4_config_get (DBusGProxy *proxy,
 	if (nm_ip4_address_get_address (addr) && nm_ip4_address_get_prefix (addr)) {
 		nm_ip4_config_take_address (config, addr);
 	} else {
-		nm_warning ("%s: invalid IP4 config received!", __func__);
+		nm_log_err (LOGD_VPN, "invalid IP4 config received!");
 		nm_ip4_address_unref (addr);
 		goto error;
 	}
@@ -456,7 +482,7 @@ nm_vpn_connection_ip4_config_get (DBusGProxy *proxy,
 		GArray *nbns = (GArray *) g_value_get_boxed (val);
 
 		for (i = 0; i < nbns->len; i++)
-			nm_ip4_config_add_nameserver (config, g_array_index (nbns, guint, i));
+			nm_ip4_config_add_wins (config, g_array_index (nbns, guint, i));
 	}
 
 	val = (GValue *) g_hash_table_lookup (config_hash, NM_VPN_PLUGIN_IP4_CONFIG_MSS);
@@ -484,35 +510,53 @@ nm_vpn_connection_ip4_config_get (DBusGProxy *proxy,
 		GSList *iter;
 
 		routes = nm_utils_ip4_routes_from_gvalue (val);
-		for (iter = routes; iter; iter = iter->next)
-			nm_ip4_config_take_route (config, (NMIP4Route *) iter->data);
+		for (iter = routes; iter; iter = iter->next) {
+			NMIP4Route *route = iter->data;
+
+			/* Ignore host routes to the VPN gateway since NM adds one itself
+			 * below.  Since NM knows more about the routing situation than
+			 * the VPN server, we want to use the NM created route instead of
+			 * whatever the server provides.
+			 */
+			if (   vpn_ext_gw
+			    && nm_ip4_route_get_dest (route) == vpn_ext_gw
+			    && nm_ip4_route_get_prefix (route) == 32)
+				continue;
+
+			/* Otherwise accept the VPN-provided route */
+			nm_ip4_config_take_route (config, route);
+		}
 
 		g_slist_free (routes);
 	}
 
-	print_vpn_config (config, priv->ip4_internal_gw, priv->tundev, priv->banner);
+	val = (GValue *) g_hash_table_lookup (config_hash, NM_VPN_PLUGIN_IP4_CONFIG_NEVER_DEFAULT);
+	if (val && G_VALUE_HOLDS_BOOLEAN (val))
+		nm_ip4_config_set_never_default (config, g_value_get_boolean (val));
+
+	print_vpn_config (config, priv->ip4_internal_gw, priv->ip_iface, priv->banner);
 
 	/* Merge in user overrides from the NMConnection's IPv4 setting */
 	s_ip4 = NM_SETTING_IP4_CONFIG (nm_connection_get_setting (priv->connection, NM_TYPE_SETTING_IP4_CONFIG));
 	nm_utils_merge_ip4_config (config, s_ip4);
 
-	nm_system_device_set_up_down_with_iface (priv->tundev, TRUE, NULL);
+	nm_system_device_set_up_down_with_iface (priv->ip_iface, TRUE, NULL);
 
-	if (nm_system_apply_ip4_config (priv->tundev, config, 0, NM_IP4_COMPARE_FLAG_ALL)) {
-		NMNamedManager *named_mgr;
+	if (nm_system_apply_ip4_config (priv->ip_iface, config, 0, NM_IP4_COMPARE_FLAG_ALL)) {
+		NMDnsManager *dns_mgr;
 
 		/* Add any explicit route to the VPN gateway through the parent device */
 		priv->gw_route = nm_system_add_ip4_vpn_gateway_route (priv->parent_dev, config);
 
 		/* Add the VPN to DNS */
-		named_mgr = nm_named_manager_get ();
-		nm_named_manager_add_ip4_config (named_mgr, priv->tundev, config, NM_NAMED_IP_CONFIG_TYPE_VPN);
-		g_object_unref (named_mgr);
+		dns_mgr = nm_dns_manager_get (NULL);
+		nm_dns_manager_add_ip4_config (dns_mgr, priv->ip_iface, config, NM_DNS_IP_CONFIG_TYPE_VPN);
+		g_object_unref (dns_mgr);
 
 		priv->ip4_config = config;
 
-		nm_info ("VPN connection '%s' (IP Config Get) complete.",
-			    nm_vpn_connection_get_name (connection));
+		nm_log_info (LOGD_VPN, "VPN connection '%s' (IP Config Get) complete.",
+		             nm_vpn_connection_get_name (connection));
 		nm_vpn_connection_set_vpn_state (connection,
 		                                 NM_VPN_CONNECTION_STATE_ACTIVATED,
 		                                 NM_VPN_CONNECTION_STATE_REASON_NONE);
@@ -520,8 +564,8 @@ nm_vpn_connection_ip4_config_get (DBusGProxy *proxy,
 	}
 
 error:
-	nm_warning ("VPN connection '%s' did not receive valid IP config information.",
-	            nm_vpn_connection_get_name (connection));
+	nm_log_warn (LOGD_VPN, "VPN connection '%s' did not receive valid IP config information.",
+	             nm_vpn_connection_get_name (connection));
 	nm_vpn_connection_set_vpn_state (connection,
 	                                 NM_VPN_CONNECTION_STATE_FAILED,
 	                                 NM_VPN_CONNECTION_STATE_REASON_IP_CONFIG_INVALID);
@@ -540,8 +584,8 @@ nm_vpn_connection_ip_config_timeout (gpointer user_data)
 	 * in this timeout, cancel activation because it's taken too long.
 	 */
 	if (nm_vpn_connection_get_vpn_state (connection) == NM_VPN_CONNECTION_STATE_IP_CONFIG_GET) {
-		nm_info ("VPN connection '%s' (IP Config Get) timeout exceeded.",
-		         nm_vpn_connection_get_name (connection));
+		nm_log_warn (LOGD_VPN, "VPN connection '%s' (IP Config Get) timeout exceeded.",
+		             nm_vpn_connection_get_name (connection));
 		nm_vpn_connection_set_vpn_state (connection,
 		                                 NM_VPN_CONNECTION_STATE_FAILED,
 		                                 NM_VPN_CONNECTION_STATE_REASON_CONNECT_TIMEOUT);
@@ -556,12 +600,12 @@ nm_vpn_connection_connect_cb (DBusGProxy *proxy, GError *err, gpointer user_data
 	NMVPNConnection *connection = NM_VPN_CONNECTION (user_data);
 	NMVPNConnectionPrivate *priv = NM_VPN_CONNECTION_GET_PRIVATE (connection);
 
-	nm_info ("VPN connection '%s' (Connect) reply received.",
-		    nm_vpn_connection_get_name (connection));
+	nm_log_info (LOGD_VPN, "VPN connection '%s' (Connect) reply received.",
+	             nm_vpn_connection_get_name (connection));
 
 	if (err) {
-		nm_warning ("VPN connection '%s' failed to connect: '%s'.", 
-				  nm_vpn_connection_get_name (connection), err->message);
+		nm_log_warn (LOGD_VPN, "VPN connection '%s' failed to connect: '%s'.", 
+		             nm_vpn_connection_get_name (connection), err->message);
 		nm_vpn_connection_set_vpn_state (connection,
 		                                 NM_VPN_CONNECTION_STATE_FAILED,
 		                                 NM_VPN_CONNECTION_STATE_REASON_SERVICE_START_FAILED);
@@ -571,7 +615,7 @@ nm_vpn_connection_connect_cb (DBusGProxy *proxy, GError *err, gpointer user_data
 		                                 NM_VPN_CONNECTION_STATE_REASON_NONE);
 		
 		/* 40 second timeout waiting for IP config signal from VPN service */
-		priv->ipconfig_timeout = g_timeout_add (40000, nm_vpn_connection_ip_config_timeout, connection);
+		priv->ipconfig_timeout = g_timeout_add_seconds (40, nm_vpn_connection_ip_config_timeout, connection);
 	}
 }
 
@@ -631,8 +675,8 @@ nm_vpn_connection_activate (NMVPNConnection *connection)
 	/* StateChanged signal */
 	dbus_g_proxy_add_signal (priv->proxy, "StateChanged", G_TYPE_UINT, G_TYPE_INVALID);
 	dbus_g_proxy_connect_signal (priv->proxy, "StateChanged",
-						    G_CALLBACK (plugin_state_changed),
-						    connection, NULL);
+	                             G_CALLBACK (plugin_state_changed),
+	                             connection, NULL);
 
 	nm_vpn_connection_set_vpn_state (connection,
 	                                 NM_VPN_CONNECTION_STATE_NEED_AUTH,
@@ -698,7 +742,7 @@ nm_vpn_connection_get_ip_iface (NMVPNConnection *connection)
 {
 	g_return_val_if_fail (NM_IS_VPN_CONNECTION (connection), NULL);
 
-	return NM_VPN_CONNECTION_GET_PRIVATE (connection)->tundev;
+	return NM_VPN_CONNECTION_GET_PRIVATE (connection)->ip_iface;
 }
 
 NMDevice *
@@ -741,147 +785,55 @@ nm_vpn_connection_disconnect (NMVPNConnection *connection,
 
 /******************************************************************************/
 
-static void
-cleanup_secrets_dbus_call (NMVPNConnection *self)
+static gboolean
+secrets_update_setting (NMSecretsProviderInterface *interface,
+                        const char *setting_name,
+                        GHashTable *new)
 {
+	NMVPNConnection *self = NM_VPN_CONNECTION (interface);
 	NMVPNConnectionPrivate *priv = NM_VPN_CONNECTION_GET_PRIVATE (self);
-	DBusGProxy *proxy;
-
-	g_return_if_fail (priv->connection != NULL);
-	g_return_if_fail (NM_IS_CONNECTION (priv->connection));
-
-	proxy = g_object_get_data (G_OBJECT (priv->connection), NM_MANAGER_CONNECTION_SECRETS_PROXY_TAG);
-	g_assert (proxy);
-
-	if (priv->secrets_call) {
-		dbus_g_proxy_cancel_call (proxy, priv->secrets_call);
-		priv->secrets_call = NULL;
-	}
-}
-
-typedef struct GetSecretsInfo {
-	NMVPNConnection *vpn_connection;
-	char *setting_name;
-} GetSecretsInfo;
-
-static void
-free_get_secrets_info (gpointer data)
-{
-	GetSecretsInfo * info = (GetSecretsInfo *) data;
-
-	g_free (info->setting_name);
-	if (info->vpn_connection)
-		g_object_unref (info->vpn_connection);
-	g_slice_free (GetSecretsInfo, info);
-}
-
-static void
-update_vpn_properties_secrets (gpointer key, gpointer data, gpointer user_data)
-{
-	NMConnection *connection = NM_CONNECTION (user_data);
-	GHashTable *secrets = (GHashTable *) data;
 	GError *error = NULL;
 
-	if (strcmp (key, NM_SETTING_VPN_SETTING_NAME))
-		return;
+	g_return_val_if_fail (priv->connection != NULL, FALSE);
 
-	if (!nm_connection_update_secrets (connection, NM_SETTING_VPN_SETTING_NAME, secrets, &error)) {
-		nm_warning ("Failed to update VPN secrets: %d %s",
-		            error ? error->code : -1,
-		            error && error->message ? error->message : "(none)");
+	if (strcmp (setting_name, NM_SETTING_VPN_SETTING_NAME))
+		return FALSE;
+
+	if (!nm_connection_update_secrets (priv->connection, NM_SETTING_VPN_SETTING_NAME, new, &error)) {
+		nm_log_warn (LOGD_VPN, "Failed to update VPN secrets: %d %s",
+		             error ? error->code : -1,
+		             error && error->message ? error->message : "(none)");
 		g_clear_error (&error);
+		return FALSE;
 	}
+	return TRUE;
 }
 
 static void
-get_secrets_cb (DBusGProxy *proxy, DBusGProxyCall *call, gpointer user_data)
+secrets_result (NMSecretsProviderInterface *interface,
+	            const char *setting_name,
+	            RequestSecretsCaller caller,
+	            const GSList *updated,
+	            GError *error)
 {
-	GetSecretsInfo *info = (GetSecretsInfo *) user_data;
-	NMVPNConnectionPrivate *priv;
-	GError *err = NULL;
-	GHashTable *settings = NULL;
+	NMVPNConnection *self = NM_VPN_CONNECTION (interface);
+	NMVPNConnectionPrivate *priv = NM_VPN_CONNECTION_GET_PRIVATE (self);
 
-	if (!info || !info->vpn_connection || !info->setting_name)
-		goto error;
+	g_return_if_fail (priv->connection != NULL);
+	g_return_if_fail (caller == SECRETS_CALLER_VPN);
 
-	priv = NM_VPN_CONNECTION_GET_PRIVATE (info->vpn_connection);
-
-	priv->secrets_call = NULL;
-
-	if (!dbus_g_proxy_end_call (proxy, call, &err,
-								DBUS_TYPE_G_MAP_OF_MAP_OF_VARIANT, &settings,
-								G_TYPE_INVALID)) {
-		nm_warning ("Couldn't get connection secrets: %s.", err->message);
-		g_error_free (err);
-		goto error;
-	}
-
-	if (g_hash_table_size (settings) == 0) {
-		// FIXME: some better way to handle invalid message?
-		nm_warning ("GetSecrets call returned but no secrets were found.");
-		goto error;
-	}
-
-	g_hash_table_foreach (settings, update_vpn_properties_secrets, priv->connection);
-	g_hash_table_destroy (settings);
-	really_activate (info->vpn_connection);
-	return;
-
-error:
-	nm_vpn_connection_fail (info->vpn_connection, NM_VPN_CONNECTION_STATE_REASON_NO_SECRETS);
+	if (error)
+		nm_vpn_connection_fail (self, NM_VPN_CONNECTION_STATE_REASON_NO_SECRETS);
+	else
+		really_activate (self);
 }
 
-static gboolean
-get_connection_secrets (NMVPNConnection *vpn_connection,
-                        const char *setting_name,
-                        gboolean request_new)
+static void
+secrets_provider_interface_init (NMSecretsProviderInterface *sp_interface_class)
 {
-	NMVPNConnectionPrivate *priv;
-	DBusGProxy *secrets_proxy;
-	GetSecretsInfo *info = NULL;
-	GPtrArray *hints;
-
-	g_return_val_if_fail (vpn_connection != NULL, FALSE);
-	g_return_val_if_fail (NM_IS_VPN_CONNECTION (vpn_connection), FALSE);
-	g_return_val_if_fail (setting_name != NULL, FALSE);
-
-	priv = NM_VPN_CONNECTION_GET_PRIVATE (vpn_connection);
-	g_assert (priv->connection);
-
-	secrets_proxy = g_object_get_data (G_OBJECT (priv->connection), NM_MANAGER_CONNECTION_SECRETS_PROXY_TAG);
-	g_assert (secrets_proxy);
-
-	info = g_slice_new0 (GetSecretsInfo);
-	g_return_val_if_fail (info != NULL, FALSE);
-
-	info->setting_name = g_strdup (setting_name);
-	info->vpn_connection = g_object_ref (vpn_connection);
-
-	/* Empty for now... */
-	hints = g_ptr_array_new ();
-
-	/* use ..._with_timeout to give the user time to enter secrets */
-	priv->secrets_call = dbus_g_proxy_begin_call_with_timeout (secrets_proxy, "GetSecrets",
-	                                                           get_secrets_cb,
-	                                                           info,
-	                                                           free_get_secrets_info,
-	                                                           G_MAXINT32,
-	                                                           G_TYPE_STRING, setting_name,
-	                                                           DBUS_TYPE_G_ARRAY_OF_STRING, hints,
-	                                                           G_TYPE_BOOLEAN, request_new,
-	                                                           G_TYPE_INVALID);
-	g_ptr_array_free (hints, TRUE);
-	if (!priv->secrets_call) {
-		nm_warning ("Could not call GetSecrets");
-		goto error;
-	}
-	return TRUE;
-
-error:
-	if (info)
-		free_get_secrets_info (info);
-	cleanup_secrets_dbus_call (vpn_connection);
-	return FALSE;
+	/* interface implementation */
+	sp_interface_class->update_setting = secrets_update_setting;
+	sp_interface_class->result = secrets_result;
 }
 
 static void
@@ -890,25 +842,32 @@ connection_need_secrets_cb  (DBusGProxy *proxy,
                              GError *error,
                              gpointer user_data)
 {
-	NMVPNConnection *vpn_connection = NM_VPN_CONNECTION (user_data);
-
-	cleanup_secrets_dbus_call (vpn_connection);
+	NMVPNConnection *self = NM_VPN_CONNECTION (user_data);
+	NMVPNConnectionPrivate *priv = NM_VPN_CONNECTION_GET_PRIVATE (self);
 
 	if (error) {
-		g_warning ("%s.%d: NeedSecrets failed: %s %s",
-		           __FILE__, __LINE__,
-		           g_quark_to_string (error->domain), error->message);
-		nm_vpn_connection_fail (vpn_connection, NM_VPN_CONNECTION_STATE_REASON_NO_SECRETS);
+		nm_log_err (LOGD_VPN, "NeedSecrets failed: %s %s",
+		            g_quark_to_string (error->domain),
+		            error->message);
+		nm_vpn_connection_fail (self, NM_VPN_CONNECTION_STATE_REASON_NO_SECRETS);
 		return;
 	}
 
-	if (setting_name && strlen (setting_name)) {
-		if (!get_connection_secrets (vpn_connection, setting_name, FALSE))
-			nm_vpn_connection_fail (vpn_connection, NM_VPN_CONNECTION_STATE_REASON_NO_SECRETS);
-	} else {
-		/* No secrets needed */
-		really_activate (vpn_connection);
+	if (!setting_name || !strlen (setting_name)) {
+		/* No secrets required */
+		really_activate (self);
+		return;
 	}
+
+	/* Get the secrets the VPN plugin wants */
+	if (!nm_secrets_provider_interface_get_secrets (NM_SECRETS_PROVIDER_INTERFACE (self),
+                                                    priv->connection,
+                                                    setting_name,
+                                                    FALSE,
+                                                    SECRETS_CALLER_VPN,
+                                                    NULL,
+                                                    NULL))
+		nm_vpn_connection_fail (self, NM_VPN_CONNECTION_STATE_REASON_NO_SECRETS);
 }
 
 static void
@@ -931,20 +890,21 @@ vpn_cleanup (NMVPNConnection *connection)
 {
 	NMVPNConnectionPrivate *priv = NM_VPN_CONNECTION_GET_PRIVATE (connection);
 
-	if (priv->tundev) {
-		nm_system_device_set_up_down_with_iface (priv->tundev, FALSE, NULL);
-		nm_system_device_flush_ip4_routes_with_iface (priv->tundev);
-		nm_system_device_flush_ip4_addresses_with_iface (priv->tundev);
+	if (priv->ip_iface) {
+		nm_system_device_set_up_down_with_iface (priv->ip_iface, FALSE, NULL);
+		/* FIXME: use AF_UNSPEC here when we have IPv6 support */
+		nm_system_device_flush_routes_with_iface (priv->ip_iface, AF_INET);
+		nm_system_device_flush_addresses_with_iface (priv->ip_iface);
 	}
 
 	if (priv->ip4_config) {
 		NMIP4Config *parent_config;
-		NMNamedManager *named_mgr;
+		NMDnsManager *dns_mgr;
 
 		/* Remove attributes of the VPN's IP4 Config */
-		named_mgr = nm_named_manager_get ();
-		nm_named_manager_remove_ip4_config (named_mgr, priv->tundev, priv->ip4_config);
-		g_object_unref (named_mgr);
+		dns_mgr = nm_dns_manager_get (NULL);
+		nm_dns_manager_remove_ip4_config (dns_mgr, priv->ip_iface, priv->ip4_config);
+		g_object_unref (dns_mgr);
 
 		/* Remove any previously added VPN gateway host route */
 		if (priv->gw_route)
@@ -957,7 +917,7 @@ vpn_cleanup (NMVPNConnection *connection)
 			                                 nm_device_get_ip4_config (priv->parent_dev),
 			                                 nm_device_get_priority (priv->parent_dev),
 			                                 NM_IP4_COMPARE_FLAG_ADDRESSES | NM_IP4_COMPARE_FLAG_ROUTES)) {
-				nm_warning ("%s: failed to re-apply VPN parent device addresses and routes.", __func__);
+				nm_log_err (LOGD_VPN, "failed to re-apply VPN parent device addresses and routes.");
 			}
 		}
 	}
@@ -972,10 +932,16 @@ vpn_cleanup (NMVPNConnection *connection)
 		priv->banner = NULL;
 	}
 
-	if (priv->tundev) {
-		g_free (priv->tundev);
-		priv->tundev = NULL;
+	if (priv->ip_iface) {
+		g_free (priv->ip_iface);
+		priv->ip_iface = NULL;
 	}
+
+	/* Clear out connection secrets to ensure that the settings service
+	 * gets asked for them next time the connection is activated.
+	 */
+	if (priv->connection)
+		nm_connection_clear_secrets (priv->connection);
 }
 
 static void
@@ -985,7 +951,7 @@ connection_state_changed (NMVPNConnection *connection,
 {
 	NMVPNConnectionPrivate *priv = NM_VPN_CONNECTION_GET_PRIVATE (connection);
 
-	cleanup_secrets_dbus_call (connection);
+	nm_secrets_provider_interface_cancel_get_secrets (NM_SECRETS_PROVIDER_INTERFACE (priv->act_request));
 
 	switch (state) {
 	case NM_VPN_CONNECTION_STATE_NEED_AUTH:
@@ -998,7 +964,7 @@ connection_state_changed (NMVPNConnection *connection,
 
 			org_freedesktop_NetworkManager_VPN_Plugin_disconnect (priv->proxy, &err);
 			if (err) {
-				nm_warning ("%s", err->message);
+				nm_log_warn (LOGD_VPN, "error disconnecting VPN: %s", err->message);
 				g_error_free (err);
 			}
 
@@ -1040,20 +1006,16 @@ dispose (GObject *object)
 	}
 	priv->disposed = TRUE;
 
-	cleanup_secrets_dbus_call (NM_VPN_CONNECTION (object));
-
 	if (priv->gw_route)
 		rtnl_route_put (priv->gw_route);
 
-	if (priv->parent_dev) {
-		if (priv->device_ip4)
-			g_signal_handler_disconnect (priv->parent_dev, priv->device_ip4);
+	if (priv->device_ip4)
+		g_signal_handler_disconnect (priv->parent_dev, priv->device_ip4);
 
-		if (priv->device_monitor)
-			g_signal_handler_disconnect (priv->parent_dev, priv->device_monitor);
+	if (priv->device_monitor)
+		g_signal_handler_disconnect (priv->parent_dev, priv->device_monitor);
 
-		g_object_unref (priv->parent_dev);
-	}
+	g_object_unref (priv->parent_dev);
 
 	if (priv->ip4_config)
 		g_object_unref (priv->ip4_config);
@@ -1064,6 +1026,7 @@ dispose (GObject *object)
 	if (priv->proxy)
 		g_object_unref (priv->proxy);
 
+	g_object_unref (priv->act_request);
 	g_object_unref (priv->connection);
 
 	G_OBJECT_CLASS (nm_vpn_connection_parent_class)->dispose (object);
@@ -1075,7 +1038,7 @@ finalize (GObject *object)
 	NMVPNConnectionPrivate *priv = NM_VPN_CONNECTION_GET_PRIVATE (object);
 
 	g_free (priv->banner);
-	g_free (priv->tundev);
+	g_free (priv->ip_iface);
 	g_free (priv->ac_path);
 
 	G_OBJECT_CLASS (nm_vpn_connection_parent_class)->finalize (object);
@@ -1106,6 +1069,9 @@ get_property (GObject *object, guint prop_id,
 	case PROP_DEFAULT:
 		g_value_set_boolean (value, priv->is_default);
 		break;
+	case PROP_DEFAULT6:
+		g_value_set_boolean (value, priv->is_default6);
+		break;
 	case PROP_VPN:
 		g_value_set_boolean (value, TRUE);
 		break;
@@ -1176,7 +1142,14 @@ nm_vpn_connection_class_init (NMVPNConnectionClass *connection_class)
 		(object_class, PROP_DEFAULT,
 		 g_param_spec_boolean (NM_ACTIVE_CONNECTION_DEFAULT,
 							   "Default",
-							   "Is the default active connection",
+							   "Is the default IPv4 active connection",
+							   FALSE,
+							   G_PARAM_READABLE));
+	g_object_class_install_property
+		(object_class, PROP_DEFAULT6,
+		 g_param_spec_boolean (NM_ACTIVE_CONNECTION_DEFAULT6,
+							   "Default6",
+							   "Is the default IPv6 active connection",
 							   FALSE,
 							   G_PARAM_READABLE));
 	g_object_class_install_property
